struct CopyTreeElem_t{
QTNode_t node;
int quadIn;
}
static 
rc_t CopyTreeToDisk(qtree_p &p, QTNode_t *node, unint4& newSlot)
{
// Temperorary variables are allocated as static
// in order to reduce the memory in the thread stack
// If slot index of the subtree is null, return
   if (node == NULL)
   {
     newSlot = SINil
     return RCOK;
   }
   static CopyTreeElem_t* 		newElem ;
   static CopyTreeElem_t* 		topElem ;
   static QTNode_t*	      		qrec_p1 ;
   qtrec_t qrec;
   qrec_p1 = NULL ; 
   newElem = topElem = NULL ;  
   uint4 si;

   w_qtstack<CopyTreeElem_t*>*	moveStack ;
   moveStack = new w_qtstack<CopyTreeElem_t*> ;
   
   w_assert3( moveStack != NULL ) ;
   moveStack->empty_all() ;
   
   qrec_p1 =  node;   /* root node of the in-memory tree */
   if ( qrec_p1->isleaf() ) /* root node is a leaf */
   {
     qrec.set_leaf();
     qrec.set_key( node->key() ) ;
     // Copy the data
     memcpy( qrec.data(), node->data(), DATA_SZ ) ;  
     goto finish ; 
   }

   /* A full tree rooted at node */
   /* Push the root into stack and then iterate */
   topElem = new CopyTreeElem_t ;
   w_assert3( topElem != NULL ) ;
   topElem->quadIn = -1 ; topElem->node = node;
   moveStack->push( topElem ) ;

   /* Iterate until stack is NOT empty */
   while ( !moveStack->is_empty() )
   {
     topElem = moveStack->pop() ;
     qrec_p1 = &(topElem->node) ;  
     if ( !qrec_p1->isleaf() )
     {
	// Find the next non null child of the current intermediate node 
	while ( ++topElem->quadIn<4&&(qrec_p1->child(topElem->quadIn)==SINil
		|| qrec_p1->ChildType(topElem->quadIn) == PTExternal) ) ;
	if ( topElem->quadIn == 4 )
	{
          qrec.set_non_leaf();
          qrec.set_key( qrec_p1->key() ) ;
          // Copy the data
          memcpy( qrec.data(), qrec_p1->data(), DATA_SZ ) ;  
	  // Copy the intermediate node
	  W_DO( CopySlot( qrec, si, p ) ) ;
	  if ( !moveStack->is_empty() )
	  {
	    delete topElem ; 
	    topElem = moveStack->pop() ;
	    qrec_p1 = &(topElem->node) ;  
	    topElem->node.child(topElem->quadIn) = si ;
	    moveStack->push(topElem) ;
	  }
	  else
	   goto clean ; 
	}
	else
	{
	  w_assert3( topElem->quadIn < 4 && 
		     qrec_p1->child(topElem->quadIn) != SINil ) ; 
	  moveStack->push( topElem ) ;
          newElem = new CopyTreeElem_t ;
	  w_assert3( newElem != NULL ) ;
	  newElem->node = (topElem->node).child(topElem->quadIn) ) ;
	  newElem->quadIn = -1 ; 
	  moveStack->push( newElem ) ;
	}
     }
     else
     {
	// Copy the node into a new slot in the page p
	si = p->AllocSlot();
        w_assert3(si != SINil);
        qrec.set_leaf();
        qrec.set_key( qrec_p1->key() ) ;
        // Copy the data
        memcpy( qrec.data(), qrec_p1->data(), DATA_SZ ) ;  
        W_DO(p.insert( qrec, newSlot ) ) ;   
	delete topElem ;
	topElem = moveStack->pop() ;
	topElem->node.child(topElem->quadIn) = si ;
	moveStack->push(topElem) ;
     }
   }
finish :
  // Copy this node to disk
  newSlot = p.AllocSlot();
  w_assert3(newSlot != SINil);
  // Insert the slot into the page
  W_DO( p.insert( qrec, newSlot ) ) ;   

clean:
  delete node ; 
  delete moveStack
  return RCOK ; 
}





struct RecCompactStackElem{
  qtrec_t slot;
  int pageNum;
  int childNum;
  QTNode_t *Qtptr;
  int childPath[4];
  int childSize[4];
};

void RecCompact(QPage_t& page, int rootIndex, QTNode_t*& qtPtr, 
                int &size, int& pathLength, int treeSizeThreshold)
{
  if ( rootIndex == SINil)
  {
    size = 0;
    pathLength = -1;
    qtPtr = NULL;
    return;
  }

  // Locals 

  w_qtstack<RecCompactStackElem_t*>*	Stack ;
  Stack = new w_qtstack<RecCompactStackElem_t*> ;  
  QPage_t &pages[200];
  int AllocPageNum = 0;
  int slotNum;
  qtrec_t slot;

  qtrec_t &rootSlot = page.slot[rootIndex];
  QTNode_t* qtNew = new QTNode_t;   // Create a new in-memory node
  Assert( qtNew != NULL );
  
  if (rootSlot.isleaf)
  {
    // Copy the leaf to a memory node
    qtNew->set_leaf();
    qtNew->set_key(rootSlot.l);
    // Free the slot occupied in the page by this leaf
    page.FreeSlot(rootIndex);

    size = 1;
    pathLength = 0; 
    qtPtr = qtNew; 
    return; 
  }

  page[AllocPageNum] = page;
  RecCompactStackElem *elem = new RecCompactStackElem;
  elem.slot = rootSlot;
  elem.pageNum = AllocPageNum;
  elem.childNum = -1;
  for (int i=0;i<4;
         elem.childPath[i]=0,elem.childSize[i]=0,qtNew->SetChildType(i,SINil),i++);
  qtNew->set_non_leaf();
  Stack->push(elem);

  while (!Stack->isempty())
  {
    RecCompactStackElem *topElem = Stack->pop();
    i = topElem->childNum;
    slot = topElem->slot;
    while (i<4)
    {
      if (slot.ChildType[i] == SINil)
      {
        i++;
        continue;
      }
    }
    if (i < 4)
    {
      if (slot.ChildType[i] == PTInternal)
      {
        qtrec_t &childSlot = pages[topElem.pageNum].rec(slot.child[i]);
        if (childSlot.isleaf())
        {
          topElem->childPath[i]=0;
          topElem->childSize[i]=1;
          QTNode_t *qtptr = new QtNode_t();
          qtptr->set_leaf();
          qtptr->set_key(childSlot.l);
          topElem->qtchild[i] = qtptr;
          pages[topElem->pageNum].FreeSlot(slot.child[i]);          
          topElem->childNum = i;
          Stack->push(topElem);
        }
        else
        {
          qtrec_t &childSlot = pages[topElem.pageNum].rec(slot.child[i]);
          RecCompactStackElem *newElem = new RecCompactStackElem();
          newElem->slot = childSlot;
          newElem->childNum = -1;
          newElem->pageNum = AllocPageNum;
          newElem->Qtptr = new QtNode_t();
          newElem->Qtptr->set_non_leaf();
          for (int i=0;i<4;newElem->Qtptr->SetChildType(i,SINil),
                           newElem->childSize[i]=0, newElem->childPath[i]=0,i++);
          topElem->ChildNum = i;
          topElem->qtchild[i] = newElem->Qtptr;
          Stack->push(topElem);
          Stack->push(newElem);
        }
      }
      else // child is external page pointer 
      {
         AllocPageNum++;
         QPage_t* p = disk.FetchPage(slot.child[q]);
         Assert(p != NULL);
         pages[AllocPageNum] = *p;
         qtrec_t &childSlot = p->rec(slot.child[i]);   
         RecCompactStackElem *newElem = new RecCompactStackElem();
         newElem->slot = childSlot;
         newElem->childNum = -1;
         newElem->pageNum = AllocPageNum;
         newElem->Qtptr = new QtNode_t();
         newElem->Qtptr->set_non_leaf();
         for (int i=0;i<4;newElem->Qtptr->SetChildType(i,SINil),
                          newElem->childSize[i]=0, newElem->childPath[i]=0,i++);
         topElem->ChildNum = i;
         topElem->qtchild[i] = newElem->Qtptr;
         Stack->push(topElem);
         Stack->push(newElem);   
       }
    }
    else
    {
      // Done with all children of topelem. So take care of this node
      DoCompact();
    }
  }
  for (int i=0;i<=AllocPageNum;i++)
   if (pages[i].usedcount == 0)
    free(pages[i]);
  
}


void DoCompact()
{
 int q, numNodes = 0;
   
  // now we have all the statistics of our children that we need.  Compute
  // the total number of nodes of all subtrees of this node.
  int numNodes = 0; 
  for (q = 0; q < 4; numNodes += childSize[q++]);    

  // Sort the trees in INCREASING order of path length 
  int max[4];
  sort(childPath, max) ;

  // decide which trees to drop and which to carry up.
  int maxPathLen = childPath[max[3]]; 
  int takeupsize = 1;               // number of nodes we are taking up
  Boolean isDropped[4] = {True, True, True, True}; // is child dropped
  
  // find the first non-null child
  for (int firstNonNull = 0; firstNonNull < 4; firstNonNull++)
    if (childPath[max[firstNonNull]] != -1)
      break;
  
  if (numNodes < treeSizeThreshold)
  {
    if (maxPathLen != childPath[max[firstNonNull]])
    {
      for (q = 3; q >=firstNonNull && childPath[max[q]]==maxPathLen; q-=1)
      {
        isDropped[max[q]] = False;
        takeupsize += childSize[max[q]];
      }
#ifndef NOTRACE
      cout << "N";
#endif
    }
    else                              
    {
      // nothing dropped. string up pointers to children and pass it up.
      for (q = 0; q < 4; q++)
      {
        qtNew->SetChildType(q, PTInternal);
        qtNew->child[q] = qtChildNodes[q];
      }
      // Free the slot occupied in the page by this node
      page.FreeSlot(rootIndex);
      size = numNodes + 1;
      pathLength = maxPathLen;
      qtPtr = qtNew;
      return;
    }
  }
  else
  {
    for (q = 3; q >=firstNonNull && childPath[max[q]]==maxPathLen; q-=1)
    {
      isDropped[max[q]] = False;
      takeupsize += childSize[max[q]];
    }
    if (takeupsize > treeSizeThreshold) // we can't take 'em up. drop all
    {
      for (q = 3; q >=firstNonNull && childPath[max[q]]==maxPathLen; q-=1)
        isDropped[max[q]] = True;
      takeupsize = 1;
    }
  }
  
  // some trees have been dropped. when we come here
#ifndef NOTRACE  
  cout.form("Decide: %d\t%d\t%d\t%d\t",
            isDropped[0],isDropped[1],isDropped[2],isDropped[3]); 
  cout.form("ChildSize: %d\t%d\t%d\t%d\t",
            childSize[0],childSize[1],childSize[2],childSize[3]); 
  cout.form("ChildPath: %d\t%d\t%d\t%d\n",
            childPath[0],childPath[1],childPath[2],childPath[3]); 
#endif
  
  // compute the path length to return
  pathLength = -1;
  for (q = 0; q < 4; q+=1)
  {
    int tmp = isDropped[q] + childPath[q];
    if (tmp > pathLength)
      pathLength = tmp;
  }

  // set the size of tree being returned.
  size  = takeupsize;
  
  // partition the children being dropped in to pages
  for (q = 0; q < 4; q+=1)
    if (!isDropped[q])
      childSize[q] = 0;             // required for partition function
  
  int assign[4] = {0, 0, 0, 0};     // assign[i] => bitset of trees in page i

  // Call Partition function that assigns the put down trees to pages optimally
  int reqPages = Partition(childSize, assign, QPage_t::SlotsPerPage);
  Assert(reqPages > 0);

  int newPages[4] = {PINil, PINil ,PINil, PINil};

  for (q = 0; q < 4; qtNew->page[q++] = PINil);
    
  for (int i = 0; i < reqPages; i+=1)
  {
    // Allocate new page
    newPages[i] = disk.AllocPage();
    Assert(newPages[i] != PINil);
    QPage_t *page = disk.FetchPage(newPages[i]);

    // flush out the nodes of the tree in memory to the current
    // iterating disk page
    for ( int q = 0; q < 4; q+=1)
    {
      // Check to which page the tree should go and make sure that the
      // tree is non empty
      if (isDropped[q] && (assign[i] & (1 << q)) && qtChildNodes[q] != NULL)
      {
        // Copy this tree from memory to the page. Watch out the memory of
        // in-memory tree is deleted once the node is copied to disk
        CopyInMemoryTreeToDiskPage(page, qtChildNodes[q], q);
        qtNew->page[q] = newPages[i] ;     
        qtNew->SetChildType(q, PTExternal) ;

        // Since the memory of qtChildNodes[q] has been deallocated
        // set it to NULL
        qtChildNodes[q] = NULL ; 

      } 
    } 
  }
    
  // Free the slot occupied in the page by this node
  page.FreeSlot(rootIndex);

  // Set the appropriate pointers for the trees to move up
  for ( q = 0 ; q < 4 ; q++ )
  {
    if (!isDropped[q])
    {
      qtNew->SetChildType(q, PTInternal);
      qtNew->child[q] = qtChildNodes[q];
    }
  }

  qtPtr = qtNew ; 

  return;
}
    
